from typing import Union
import datetime
import numpy as np
import pandas as pd
from paux import date as _date
from paux.param import isnull
from candlelite import exception
from candlelite.calculate import interval as _interval

__all__ = ['valid_interval', 'valid_start', 'valid_end', 'valid_length']


# 验证数据间隔
def valid_interval(
        candle: np.array,
        interval: Union[int, float] = None,
        bar: str = None,
        MINUTE_BAR_INTERVAL: int = 60000
)->dict:
    '''
    :param candle: 历史K线数据
    :param interval: 间隔
    :param bar: 粒度
    :param MINUTE_BAR_INTERVAL: 每分钟的间隔单位(毫秒)
    :return:
        {
            'code' : True|False,    # True 验证成功 False 验证失败
            'data' : {},            # 空字典可忽略
            'msg'  : '...'          # 验证失败时记录失败原因
        }
    '''
    # 空数据
    if isnull(candle):
        return {'code': False, 'data': {}, 'msg': '[candle empty]'}
    # interval和bar不能同时为空
    if not interval and not bar:
        msg = 'Interval and bar cannot be None at the same time'
        raise exception.ParamException(func='valid_interval', msg=msg)
    # 优先级interval >> bar
    if pd.isnull(interval):
        interval = _interval.get_interval(bar=bar, MINUTE_BAR_INTERVAL=MINUTE_BAR_INTERVAL)
    # 时间间隔
    diffs = np.diff(candle[:, 0])
    if (diffs == interval).all():
        result = {'code': True, 'data': {}, 'msg': ''}
    else:
        diffs_unique = ','.join(list(set(diffs.tolist())))
        msg = '[valid candle interval error]: correct_interval={correct_interval} error_interval={error_interval}'.format(
            correct_interval=interval,
            error_interval=diffs_unique,
        )
        result = {'code': False, 'data': {}, 'msg': msg}
    return result


# 验证数据的起始时间
def valid_start(
        candle: np.array,
        start: Union[int, float, str, datetime.date],
        timezone: str = None
)->dict:
    '''
    :param candle: 历史K线数据
    :param start: 起点
        支持 毫秒级时间戳、日期格式字符串、日期类型与日期时间类型
    :param timezone: 时区
        {
            'code' : True|False,    # True 验证成功 False 验证失败
            'data' : {},            # 空字典可忽略
            'msg'  : '...'          # 验证失败时记录失败原因
        }
    '''
    if isnull(candle):
        result = {'code': False, 'data': {}, 'msg': '[candle empty]'}
        return result
    # 起始毫秒时间戳
    ts = _date.to_ts(date=start, timezone=timezone, default=np.nan)
    if candle[0, 0] == ts:
        result = {'code': True, 'data': {}, 'msg': ''}
    else:
        msg = '[valid candle start error]: correct_ts={correct_ts} error_ts={error_ts}'.format(
            correct_ts=str(start),
            error_ts=str(candle[0, 0]),
        )
        result = {'code': False, 'data': {}, 'msg': msg}
    return result


# 验证数据的终止时间
def valid_end(
        candle: np.array,
        end: Union[int, float, str, datetime.date],
        timezone: str = None
)->dict:
    '''
    :param candle: 历史K线数据
    :param start: 终点
        支持 毫秒级时间戳、日期格式字符串、日期类型与日期时间类型
    :param timezone: 时区
        {
            'code' : True|False,    # True 验证成功 False 验证失败
            'data' : {},            # 空字典可忽略
            'msg'  : '...'          # 验证失败时记录失败原因
        }
    '''
    # 终点毫秒时间戳
    if isnull(candle):
        result = {'code': False, 'data': {}, 'msg': '[candle empty]'}
        return result
    ts = _date.to_ts(date=end, timezone=timezone, default=np.nan)
    if candle[-1, 0] == ts:
        result = {'code': True, 'data': {}, 'msg': ''}
    else:
        msg = '[valid candle end error]: correct_ts={correct_ts} error_ts={error_ts}'.format(
            correct_ts=str(end),
            error_ts=str(candle[-1, 0]),
        )
        result = {'code': False, 'data': {}, 'msg': msg}
    return result


def valid_length(
        candle: np.array,
        length: int,
)->dict:
    '''
    :param candle: 历史K线数据
    :param length: 长度
            {
            'code' : True|False,    # True 验证成功 False 验证失败
            'data' : {},            # 空字典可忽略
            'msg'  : '...'          # 验证失败时记录失败原因
        }
    '''
    if isnull(candle):
        result = {'code': False, 'data': {}, 'msg': '[candle empty]'}
        return result
    if candle.shape[0] != length:
        result = {
            'code': False,
            'data': {},
            'msg': '[valid candle length error]: correct_length={correct_length} error_length={error_length}'.format(
                correct_length=length,
                error_length=candle.shape[0]
            )
        }
    else:
        result = {
            'code': True,
            'data': {},
            'msg': ''
        }
    return result



from typing import Union, Literal
import os
import numpy as np
import pandas as pd
import datetime
from paux import param as _param
from paux import process as _process
from paux import date as _date
from candlelite.calculate import transform as _transform
from candlelite.calculate import valid as _valid
from candlelite.calculate import interval as _interval
from candlelite.io import path as _path
from candlelite import exception

__all__ = [
    'load_candle_by_date',
    'load_candle_by_file',
    'load_candle_map_by_date',
    'load_candle_map_by_file',
]


# 读取从start~end日期的历史K线数据
def load_candle_by_date(
        instType: str,
        symbol: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
) -> np.ndarray:
    '''
    :param instType: 产品类型
    :param symbol: 产品名称
    :param start: 起始时间

    :param end: 终止时间
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :param valid_interval: 是否验证数据时间间隔
    :param valid_start: 是否验证数据起始时间
    :param valid_end: 是否验证数据终止时间
    '''
    # 文件是否存在
    check_result = _path.check_candle_date_path(
        instType=instType,
        symbol=symbol,
        start=start,
        end=end,
        timezone=timezone,
        bar=bar,
        base_dir=base_dir,
    )
    if not check_result['code']:
        raise exception.CandleFileNotExist(
            symbol=symbol,
            date=str([data['date'] for data in check_result['data']]),
            path=None,

        )
    # 日期序列
    date_range = _date.get_range_dates(
        start=start,
        end=end,
        timezone=timezone,
    )
    # 文件路径
    paths = [
        _path.get_candle_date_path(
            instType=instType,
            symbol=symbol,
            date=date,
            timezone=timezone,
            bar=bar,
            base_dir=base_dir,
        )
        for date in date_range
    ]
    # 读取->DataFrame
    dfs = [pd.read_csv(path) for path in paths]
    # 合并数据->Candle
    candle = _transform.concat_candle(candles=dfs, drop_duplicate=True, sort=True)
    # 验证interval
    if valid_interval:
        valid_interval_result = _valid.valid_interval(candle=candle, bar=bar)
        if not valid_interval_result['code']:
            raise exception.CandleIntervalError(
                symbol=symbol,
                msg=valid_interval_result['msg']
            )
    # 验证start
    if valid_start:
        start_ts = _date.to_ts(date=date_range[0], timezone=timezone)
        valid_start_result = _valid.valid_start(candle=candle, start=start_ts, timezone=timezone)
        if not valid_start_result['code']:
            raise exception.CandleStartError(
                symbol=symbol,
                msg=valid_start_result['msg'],
            )
    # 验证end
    if valid_end:
        end_ts = _date.tomorrow(date=date_range[-1],timezone=timezone).timestamp() * 1000 - _interval.get_interval(bar)
        # end_ts = _date.to_ts(date=date_range[-1], timezone=timezone) + 1000 * 60 * 60 * 24 - _interval.get_interval(bar)
        valid_end_result = _valid.valid_end(candle=candle, end=end_ts, timezone=timezone)
        if not valid_end_result['code']:
            raise exception.CandleEndError(
                symbol=symbol,
                msg=valid_end_result['msg'],
            )
    return candle


# 按照日期读取candle_map
def load_candle_map_by_date(
        instType: str,
        symbols: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        endswith: str = '',
        contains:str = '',
        p_num: int = 4,
        valid_interval: bool = True,
        valid_start: bool = True,
        valid_end: bool = True,
) -> dict:
    '''
    :param instType: 产品类型
    :param symbols: 产品名称列表
    :param start: 起始时间
    :param end: 终止时间
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :param endswith: 产品名称需以此结尾
    :param contains: 产品名称需包含此内容
    :param valid_interval: 是否验证数据时间间隔
    :param valid_start: 是否验证数据起始时间
    :param valid_end: 是否验证数据终止时间
    '''
    # 如果没有产品的名字，获取产品类型数据中，有start_date到end_date中有完整数据的symbol
    if not symbols:
        dates = _date.get_range_dates(start=start, end=end, timezone=timezone)
        symbols = set()
        for date in dates:
            date_dirpath = os.path.dirname(
                _path.get_candle_date_path(
                    base_dir=base_dir,
                    instType=instType,
                    timezone=timezone,
                    date=date,
                    bar=bar,
                    symbol=''
                )
            )
            this_symbols = []
            for filename in os.listdir(date_dirpath):
                symbol = filename.rsplit('.', maxsplit=1)[0]
                if endswith and not symbol.endswith(endswith):
                    continue
                if contains and not contains in symbol:
                    continue
                else:
                    this_symbols.append(symbol)
            if not symbols:
                symbols = set(this_symbols)
            else:
                symbols = symbols & set(this_symbols)
    symbols = list(symbols)

    candle_map = {}
    if p_num > 1:
        params = []
        for symbol in symbols:
            params.append(
                dict(
                    instType=instType,
                    symbol=symbol,
                    start=start,
                    end=end,
                    timezone=timezone,
                    bar=bar,
                    base_dir=base_dir,
                    valid_interval=valid_interval,
                    valid_start=valid_start,
                    valid_end=valid_end,
                )
            )
        results = _process.pool_worker(
            params=params,
            p_num=p_num,
            func=load_candle_by_date,
            skip_exception=False,
        )
        for i, candle in enumerate(results):
            if _param.isnull(candle):
                continue
            symbol = symbols[i]
            candle_map[symbol] = candle

    else:
        for symbol in symbols:
            candle_map[symbol] = load_candle_by_date(
                instType=instType,
                symbol=symbol,
                start=start,
                end=end,
                timezone=timezone,
                bar=bar,
                base_dir=base_dir,
                valid_interval=valid_interval,
                valid_start=valid_start,
                valid_end=valid_end,
            )
    # candle_map排序
    candle_map_sorted = {}
    symbols = sorted([symbol for symbol in candle_map.keys()])
    for symbol in symbols:
        candle_map_sorted[symbol] = candle_map[symbol]
    return candle_map_sorted


# 通过文件地址读取Candle
def load_candle_by_file(
        instType: str,
        symbol: str,
        path: str = None,
        base_dir: str = '',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,
):
    '''
    如果有path路径，按照path路径读取文件
    如果没有path路径，按照base_dir、symbol、instType、bar和timezone计算产品路径
    '''
    # 路径
    if path == None:
        path = _path.get_candle_file_path(
            instType=instType,
            symbol=symbol,
            bar=bar,
            timezone=timezone,
            base_dir=base_dir,

        )
    # 读取
    df = pd.read_csv(path)
    candle = _transform.to_candle(candle=df, drop_duplicate=True, sort=True)
    # 验证interval
    if valid_interval:
        valid_interval_result = _valid.valid_interval(candle=candle, bar=bar)
        if not valid_interval_result['code']:
            raise exception.CandleIntervalError(
                symbol=symbol,
                msg=valid_interval_result['msg']
            )
    return candle


# 通过文件夹地址读取Candle_map
def load_candle_map_by_file(
        instType: str,
        symbols: list = [],
        path: str = None,
        base_dir: str = '',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval: bool = True,
):
    # 路径
    if path == None:
        path = os.path.dirname(
            _path.get_candle_file_path(
                instType=instType,
                base_dir=base_dir,
                timezone=timezone,
                bar=bar,
                symbol='xx',
            )
        )
    # symbols
    if not symbols:
        filenames = os.listdir(path)
        symbols = []
        for filename in filenames:
            symbol = filename.rsplit('.', maxsplit=1)[0]
            symbols.append(symbol)
    # 读取数据
    candle_map = {}
    for symbol in symbols:
        candle = load_candle_by_file(
            symbol=symbol,
            instType=instType,
            base_dir=base_dir,
            timezone=timezone,
            bar=bar,
            valid_interval=valid_interval
        )
        candle_map[symbol] = candle
    # candle_map排序
    candle_map_sorted = {}
    symbols = sorted([symbol for symbol in candle_map.keys()])
    for symbol in symbols:
        candle_map_sorted[symbol] = candle_map[symbol]
    return candle_map_sorted



from typing import Literal, Union
import os
import re
import datetime
from paux import date as _date

__all__ = [
    'get_candle_date_path',  # 获取某一个天candle的路径
    'get_candle_file_path',  # 获取candle文件的地址（一般不以天切割，必须缓存数据与1d数据可以储存在一个文件中）
    'check_candle_date_path',  # 检查candle文件是否存在（不验证数据的准确性）
    'check_candle_file_path',  # 检查candle从start到end日期数据文件是否齐全（仅检查文件是否存在，并不验证文件的准确性）
]


# 将instType、timezone与bar转换成文件夹的名字
def _get_date_dirname(
        instType: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m'
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param bar: 时间粒度
    :return: 文件夹的名字（并不是文件夹的路径）
    '''
    if timezone == None:
        timezone = ''
    timezone = re.sub('/', '', timezone)
    return '-'.join([instType, timezone, bar])


# 将instType、timezone与bar转换成文件夹的名字
def _get_file_dirname(
        instType: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m'
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param bar: 时间粒度
    :return: 文件夹的名字（并不是文件夹的路径）
    '''
    if timezone == None:
        timezone = ''
    timezone = re.sub('/', '', timezone)
    return '-'.join([instType, timezone, bar, 'FILE'])


# 获取某一个天candle的路径
def get_candle_date_path(
        instType: str,
        symbol: str,
        date: datetime.date,
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param date: 日期
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :return: 某产品在指定日期的candle数据路径
    '''
    FMT = '%Y-%m-%d'
    date_str = _date.to_fmt(date=date, timezone=timezone, fmt=FMT)
    filepath = os.path.join(
        base_dir,
        _get_date_dirname(instType=instType, timezone=timezone, bar=bar),
        date_str[0:7],
        date_str,
        '%s.csv' % symbol
    )
    return filepath


# 获取candle文件的地址（一般不以天切割，必须缓存数据与1d数据可以储存在一个文件中）
def get_candle_file_path(
        instType: str,
        symbol: str,
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :return: candle文件的路径
    '''
    FMT = '{symbol}.csv'
    filepath = os.path.join(
        base_dir,
        _get_file_dirname(instType=instType, timezone=timezone, bar=bar),
        FMT.format(symbol=symbol)
    )
    return filepath


# 检查candle文件是否存在（不验证数据的准确性）
def check_candle_file_path(
        instType: str,
        symbol: str,
        base_dir: str,
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :return:
        True    有文件
        False   无文件
    '''
    path = get_candle_file_path(
        instType=instType,
        symbol=symbol,
        base_dir=base_dir,
        timezone=timezone,
        bar=bar
    )
    result = {
        'code': os.path.isfile(path),
        'data': path,
        'msg': path,
    }
    return result


# 检查candle从start到end日期数据文件是否齐全（仅检查文件是否存在，并不验证文件的准确性）
def check_candle_date_path(
        instType: str,
        symbol: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir:str ='',
        timezone: str = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
):
    '''
    :param instType: 产品类别
    :param symbol: 产品名称
    :param start: 起始日期
    :param end: 终止日期（包含）
    :param base_dir: 数据文件夹
    :param timezone: 时区
    :param bar: 时间粒度
    :return:
        code:
            True    数据齐全
            False   数据不全
    '''
    dates = _date.get_range_dates(start=start, end=end, timezone=timezone)
    result = {'code': True, 'data': [], 'msg': ''}  # data保存不存在数据的日期与路径

    for date in sorted(dates, reverse=True):
        path = get_candle_date_path(
            instType=instType, symbol=symbol, date=date,
            timezone=timezone, base_dir=base_dir, bar=bar
        )
        if not os.path.isfile(path):
            result['code'] = False
            result['data'].append(
                {
                    'date': date,
                    'path': path,
                }
            )
    return result





'''
compress_candle             压缩历史K线
extract_candle              根据时间跨度截取K线
concat_candle               合并数据
to_candle                   转换为ndarray类型的K数据
get_candle_index_by_date    根据日期时间，得到K线中的行索引
'''

from typing import Union
import datetime
import numpy as np
import pandas as pd
from paux import date as _date
from candlelite import exception
from candlelite.calculate import bar as _bar
from candlelite.calculate import interval as _interval

__all__ = ['compress_candle', 'extract_candle', 'to_candle', 'concat_candle', 'get_candle_index_by_date']


# 压缩历史K线
def compress_candle(
        candle: np.array,
        target_bar: str,
        org_bar: str = 'auto'
) -> np.ndarray:
    '''
    :param candle: 历史K线数据
    :param target_bar: 目标K线的bar
    :param org_bar: 原始K线的bar
       auto: 自动识别原始K线的bar
    :return:压缩后的历史K线数据
        array([
            [ts,open,high,low,close,volume...],
            [ts,open,high,low,close,volume...],
            [ts,open,high,low,close,volume...],
        ])
    example：
        将1Minute的K线数据压缩成5Minute的K线数据
        compress_candle(candle,target_bar='5m',org_bar='1m')

        将1Hour的K线数据压缩成1Day的K线数据
        compress_candle(candle,target_bar='1d',org_bar='1h'
    '''
    if org_bar == 'auto':
        org_bar = _bar.predict_bar(candle)
    # 目标K线ts间隔
    target_bar_interval = _interval.get_interval(target_bar)
    # 原始K线ts间隔
    org_tar_interval = _interval.get_interval(org_bar)
    # 压缩的数量
    compress_quantity = target_bar_interval / org_tar_interval
    if compress_quantity != int(compress_quantity):
        raise exception.ExecuteException(
            func='compress_candle',
            msg="Can't transform candle from org_bar=({org_bar}) to target_bar={target_bar}".format(
                org_bar=org_bar,
                target_bar=target_bar,
            )
        )
    compress_quantity = int(compress_quantity)
    # 目标K线数据
    target_datas = []
    candle_shape = candle.shape
    for i in range(0, candle.shape[0], compress_quantity):
        if i + compress_quantity <= candle_shape[0]:
            this_data = [
                candle[i, 0],  # ts
                candle[i, 1],  # open
                candle[i:i + compress_quantity, 2].max(),  # high
                candle[i:i + compress_quantity, 3].min(),  # low
                candle[i + compress_quantity - 1, 4],  # close
                candle[i:i + compress_quantity, 5].sum()  # volume
            ]
            # 如果有其他数据
            for i in range(6, candle_shape[1]):
                this_data.append(
                    candle[i:i + compress_quantity, i].sum()
                )
            target_datas.append(this_data)
    # 目标K线Candle
    target_candle = np.array(target_datas)
    return target_candle


# 根据时间跨度截取candle
def extract_candle(
        candle: np.array,
        start: Union[int, float, str, datetime.date, None],
        end: Union[int, float, str, datetime.date, None],
        timezone: Union[str, None] = None,
) -> np.ndarray:
    '''
    :param candle: 历史K线数据
    :param start: 数据起点 (包含起点）
    :param end: 数据终点 (包含终点）
    :param timezone: 时区
    '''
    start_ts = _date.to_ts(date=start, timezone=timezone, default=0)
    end_ts = _date.to_ts(date=end, timezone=timezone, default=candle[:, 0].max())
    return candle[(candle[:, 0] >= start_ts) & (candle[:, 0] <= end_ts)]


# 转换为candle数据
def to_candle(
        candle: Union[list, tuple, np.ndarray, pd.DataFrame],
        drop_duplicate: bool = True,
        sort: bool = True
) -> np.ndarray:
    '''
    :param candle: 历史K线数据，支持列表、元组、array、DataFrame
    :param drop_duplicate: 去重
    :param sort: 排序
    '''
    # list和tuple
    if isinstance(candle, list) or isinstance(candle, tuple):
        df = pd.DataFrame(candle)
    # DataFrame
    elif isinstance(candle, pd.DataFrame):
        df = candle
    # Array
    elif isinstance(candle, np.ndarray):
        df = pd.DataFrame(candle)
    # 未知类型
    else:
        raise exception.ParamException(
            func='to_candle',
            msg='input candle type is {candle_type}, candle type must in [list,tuple,pd.DataFrame,np.ndarray]'.format(
                candle_type=type(candle).__name__
            )
        )
    # 时间序列的名字
    ts_column_name = df.columns[0]
    # 时间戳转化为整数
    df[ts_column_name] = df[ts_column_name].astype(int)
    # 去重
    if drop_duplicate:
        df = df.drop_duplicates(subset=ts_column_name)
    # 排序
    if sort:
        df = df.sort_values(by=ts_column_name)
    # 转化为array对象
    candle = df.to_numpy()
    return candle


# 合并数据
def concat_candle(
        candles: list,
        drop_duplicate: bool = True,
        sort: bool = True
) -> np.ndarray:
    '''
    :param candles: 多个历史K线数据
    :param drop_duplicate: 去重
    :param sort: 排序
    '''
    for i in range(len(candles)):
        # list和tuple
        if isinstance(candles[i], list) or isinstance(candles[i], tuple):
            candles[i] = pd.DataFrame(candles[i])
        # DataFrame
        elif isinstance(candles[i], pd.DataFrame):
            pass
        # Array
        elif isinstance(candles[i], np.ndarray):
            candles[i] = pd.DataFrame(candles[i])
        # 未知类型
        else:
            raise exception.ParamException(
                func='concat_candle',
                msg='input candle type is {candle_type}, candle type must in [list,tuple,pd.DataFrame,np.ndarray]'.format(
                    candle_type=type(candles[i]).__name__
                ),
            )
    # 拼接数据
    df = pd.concat(candles)
    candle = to_candle(candle=df, drop_duplicate=drop_duplicate, sort=sort)
    return candle


# 根据日期时间，得到K线中的行索引
def get_candle_index_by_date(
        candle: np.array,
        date: Union[datetime.datetime, int, float, str,],
        timezone: str = None,
        default: int = 0
) -> int:
    '''
    :param candle: 历史K线数据
    :param date: 日期时间
    :param timezone: 时区
    :param default: 默认值
    :return: 航索引
    '''
    if not date:
        return default
    ts = _date.to_ts(date=date, timezone=timezone, default=default)
    index = np.where(
        candle[:, 0] == ts
    )[0][0]
    return index



import os

__all__ = [
    'get_settings_filepath',
    'save_settings',
    'init_settings',
    'read_settings',
    'show_settings',
    'console_settings'
]
# 配置文件编码，解决跨平台问题
ENCODING = 'UTF-8'
# 初始化的配置文件内容 key:[value,note]
INIT_DATA = {
    "CANDLE_BASE_DIR": ["'CANDLELITE_DATA'", '历史K线数据根目录'],

    "OKX_DATE_DIRNAME": ["'OKX'", 'OKX以日期为单位的存储目录'],
    "OKX_FILE_DIRNAME": ["'OKX_FILE'", 'OKX以文件为单位的存储目录'],
    "OKX_TIMEZONE": ["'Asia/Shanghai'", 'OKX的默认时区'],
    "OKX_DEFAULT_BAR": ["'1m'", 'OKX的默认时间粒度'],

    "BINANCE_DATE_DIRNAME": ["'BINANCE'", 'BINANCE以日期为单位的存储目录'],
    "BINANCE_FILE_DIRNAME": ["'BINANCE_FILE'", 'BINANCE以文件为单位的存储目录'],
    "BINANCE_TIMEZONE": ["'America/New_York'", 'BINANCE的默认时区'],
    "BINANCE_DEFAULT_BAR": ["'1m'", 'BINANCE的默认时间粒度'],

}


# 获取配置文件路径
def get_settings_filepath(filename: str = 'SETTINGS.config'):
    '''
    :param filename: 配置文件的名字
    :return: 配置文件路径
    '''
    dirpath = os.path.dirname(os.path.dirname(__file__))
    settings_path = os.path.join(dirpath, filename)
    return settings_path


# 保存到配置文件
def save_settings(data: dict, settings_path: str = None) -> None:
    content = ''
    for key, (value, note) in data.items():
        line = '# {note}\n{key} = {value}\n'.format(
            note=note,
            key=key,
            value=value,
        )
        content += line
    content = content.strip()
    if not settings_path:
        settings_path = get_settings_filepath()
    with open(settings_path, 'w', encoding=ENCODING) as f:
        f.write(content)


# 初始化配置文件
def init_settings(settings_path: str = None) -> None:
    save_settings(data=INIT_DATA, settings_path=settings_path)
    print('init settings complete')


# 读取配置文件
def read_settings(settings_path: str = None) -> dict:
    '''
    :param settings_path: 配置文件路径，不填写用默认
    :return: 配置文件字典 如果没有配置文件，会返回空字典
    '''
    if not settings_path:
        settings_path = get_settings_filepath()
    # 无文件 初始化文件
    if not os.path.isfile(settings_path):
        init_settings()
    with open(settings_path, 'r', encoding=ENCODING) as f:
        content = f.read().strip()
        content_list = content.split('#')
    data = {}
    for line in content_list:
        line = line.strip()
        if not line:
            continue
        note, kv = line.split('\n')
        k, v = kv.split('=')
        note = note.strip()
        k = k.strip()
        v = v.strip()
        data[k] = [v, note]
    return data


# 展示配置文件内容
def show_settings(settings_path: str = None) -> None:
    if not settings_path:
        settings_path = get_settings_filepath()
    msg = f'settings_path={settings_path}\nYou can use a text editor to modify the settings file, or you can use candlelite.settings.console_settings to modify\n'
    print(msg)
    with open(settings_path, 'r', encoding=ENCODING) as f:
        content = f.read()
        print(content)


# 在终端中修改配置文件
def console_settings(settings_path: str = None) -> None:
    '''
    :param settings_path: 配置文件路径，不填写用默认
    '''
    setting = read_settings(settings_path)
    for keyword in setting.keys():
        while True:
            value = setting[keyword][0]
            note = setting[keyword][1]
            msg = f'#{note}\n{keyword}={value} (Y/N)'
            command = input(msg)
            if command.upper() == 'N':
                msg = f'{keyword}='
                command = input(msg).strip()
                # 转化为字符串格式，收尾添加引号
                if not command.startswith("'"):
                    command = "'" + command
                if not command.endswith("'"):
                    command = command + "'"
                setting[keyword][0] = command
                break
            elif command.upper() == 'Y':
                break
            else:
                msg = 'please enter Y (YES) or N (NO)'
                print(msg)
    save_settings(data=setting, settings_path=settings_path)



import os
from candlelite.settings import read_settings
from candlelite.crypto._base import IO

settings = read_settings()
CANDLE_BASE_DIR = settings['CANDLE_BASE_DIR'][0]  # 根目录
__BINANCE_DATE_DIRNAME = settings['BINANCE_DATE_DIRNAME'][0]
__BINANCE_FILE_DIRNAME = settings['BINANCE_FILE_DIRNAME'][0]
# 以日期为单位的存储目录
BINANCE_CANDLE_DATE_BASE_DIR = os.path.join(CANDLE_BASE_DIR, __BINANCE_DATE_DIRNAME)
# 以文件为单位的存储目录
BINANCE_CANDLE_FILE_BASE_DIR = os.path.join(CANDLE_BASE_DIR, __BINANCE_FILE_DIRNAME)
# 时区
BINANCE_TIMEZONE = settings['BINANCE_TIMEZONE'][0]
# 默认时间粒度
BINANCE_DEFAULT_BAR = settings['BINANCE_DEFAULT_BAR'][0]


class BinanceLite(IO):
    CANDLE_DATE_BASE_DIR = BINANCE_CANDLE_DATE_BASE_DIR
    CANDLE_FILE_BASE_DIR = BINANCE_CANDLE_FILE_BASE_DIR
    TIMEZONE = BINANCE_TIMEZONE
    BAR = BINANCE_DEFAULT_BAR



from candlelite.exception._base import AbstractEXP, get_error_msg


class ParamException(AbstractEXP):
    def __init__(self, func, msg):
        self.error_msg = get_error_msg(func=func, reason=msg)


class ParamBarException(AbstractEXP):
    def __init__(self, func, bar, pmt="bar must like ['1m','3m','5m','30m','1h','2h','4h','6h','1d'...]"):
        reason = 'bar={bar}\n{pmt}'.format(
            bar=str(bar),
            pmt=pmt,
        )
        self.error_msg = get_error_msg(func=func, reason=reason)


class PosSideException(AbstractEXP):
    def __init__(self, func, posSide, pmt="posSide must in ['LONG','SHORT']"):
        reason = 'posSide={posSide}\n{pmt}'.format(
            posSide=str(posSide),
            pmt=pmt,
        )
        self.error_msg = get_error_msg(func=func, reason=reason)



from candlelite.exception._base import AbstractEXP


class CandleFileNotExist(AbstractEXP):
    def __init__(
            self,
            symbol: str,
            date: str = None,
            path: str = None,
    ):
        self.error_msg = 'symbol={symbol:<10} date={date} path={path}'.format(
            symbol=str(symbol),
            date=str(date),
            path=str(path),
        )


class CandleIntervalError(AbstractEXP):
    def __init__(
            self,
            symbol: str,
            msg: str,
    ):
        self.error_msg = 'symbol={symbol:<10} msg={msg}'.format(
            symbol=str(symbol),
            msg=str(msg)
        )


class CandleStartError(AbstractEXP):
    def __init__(
            self,
            symbol: str,
            msg: str,
    ):
        self.error_msg = 'symbol={symbol:<10} msg={msg}'.format(
            symbol=str(symbol),
            msg=str(msg)
        )


class CandleEndError(AbstractEXP):

    def __init__(
            self,
            symbol: str,
            msg: str,
    ):
        self.error_msg = 'symbol={symbol:<10} msg={msg}'.format(
            symbol=str(symbol),
            msg=str(msg)
        )
